pro loadtxt,FILE,OUT,M
;+
; NAME:
;       LOADTXT
;
; PURPOSE:
;       Read an ASCII file with multiple columns generated by GHOST, 
;       with a comand similar to load in MATLAB and to loadtxt in numpy.
;
; CALLING SEQUENCE:
;       LOADTXT,FILE,OUT
;
; INPUTS:
;       FILE:   A string containing the file name.
;
;       M:      An integer with the number of columns in the file 
;               (optional). If M is not set, M=1 is assumed.
;
; OUTPUTS:
;       OUT:    An array with the contents of the file.
;
; KEYWORD PARAMETERS:
;       None.
;
; COMMON BLOCKS:
;       None.
;
; AUTHOR:
;       Pablo Daniel Mininni
;-

if keyword_set(M) eq 0 then M = 1

OUX = fltarr(M,5000000)
aux1 = double(0)
aux2 = double(0)
aux3 = double(0)
aux4 = double(0)
aux5 = double(0)
N = 0L

openr,1,FILE

if M EQ 1 then begin
   while not eof(1) do begin 
      readf,1,aux1
      OUX(0,N) = aux1
      N = N+1
   endwhile
endif
if M EQ 2 then begin
   while not eof(1) do begin 
      readf,1,aux1,aux2
      OUX(0,N) = aux1
      OUX(1,N) = aux2
      N = N+1
   endwhile
endif
if M EQ 3 then begin
   while not eof(1) do begin 
      readf,1,aux1,aux2,aux3
      OUX(0,N) = aux1
      OUX(1,N) = aux2
      OUX(2,N) = aux3
      N = N+1
   endwhile
endif
if M EQ 4 then begin
   while not eof(1) do begin 
      readf,1,aux1,aux2,aux3,aux4
      OUX(0,N) = aux1
      OUX(1,N) = aux2
      OUX(2,N) = aux3
      OUX(3,N) = aux4
      N = N+1
   endwhile
endif
if M EQ 5 then begin
   while not eof(1) do begin
      readf,1,aux1,aux2,aux3,aux4,aux5
      OUX(0,N) = aux1
      OUX(1,N) = aux2
      OUX(2,N) = aux3
      OUX(3,N) = aux4
      OUX(4,N) = aux5
      N = N+1
   endwhile
endif

close,1

OUT = fltarr(M,N)
OUT = OUX(0:M-1,0:N-1)

end
